@@ -1,22 +1,13 @@
-#!/usr/bin/env python
-# -*- coding: utf-8 -*-
-
-from __future__ import absolute_import
-from __future__ import division
-from __future__ import print_function
-from __future__ import unicode_literals
-
-__all__ = ['singledispatch']
+__all__ = ['singledispatch', 'singledispatchmethod']
 
+from abc import get_cache_token
 from functools import update_wrapper
-from weakref import WeakKeyDictionary
-
-from .helpers import MappingProxyType, get_cache_token, get_type_hints
 
 ################################################################################
-### singledispatch() - single-dispatch generic function decorator
+# singledispatch() - single-dispatch generic function decorator
 ################################################################################
 
+
 def _c3_merge(sequences):
     """Merges MROs in *sequences* to a single MRO using the C3 algorithm.
 
@@ -25,18 +16,18 @@ def _c3_merge(sequences):
     """
     result = []
     while True:
-        sequences = [s for s in sequences if s]   # purge empty sequences
+        sequences = [s for s in sequences if s]  # purge empty sequences
         if not sequences:
             return result
-        for s1 in sequences:   # find merge candidates among seq heads
+        for s1 in sequences:  # find merge candidates among seq heads
             candidate = s1[0]
             for s2 in sequences:
                 if candidate in s2[1:]:
                     candidate = None
-                    break      # reject the current head, it appears later
+                    break  # reject the current head, it appears later
             else:
                 break
-        if not candidate:
+        if candidate is None:
             raise RuntimeError("Inconsistent hierarchy")
         result.append(candidate)
         # remove the chosen candidate
@@ -44,6 +35,7 @@ def _c3_merge(sequences):
             if seq[0] == candidate:
                 del seq[0]
 
+
 def _c3_mro(cls, abcs=None):
     """Computes the method resolution order using extended C3 linearization.
 
@@ -64,7 +56,7 @@ def _c3_mro(cls, abcs=None):
     for i, base in enumerate(reversed(cls.__bases__)):
         if hasattr(base, '__abstractmethods__'):
             boundary = len(cls.__bases__) - i
-            break   # Bases up to the last explicit ABC are considered first.
+            break  # Bases up to the last explicit ABC are considered first.
     else:
         boundary = 0
     abcs = list(abcs) if abcs else []
@@ -73,8 +65,8 @@ def _c3_mro(cls, abcs=None):
     other_bases = list(cls.__bases__[boundary:])
     for base in abcs:
         if issubclass(cls, base) and not any(
-                issubclass(b, base) for b in cls.__bases__
-            ):
+            issubclass(b, base) for b in cls.__bases__
+        ):
             # If *cls* is the class that introduces behaviour described by
             # an ABC *base*, insert said ABC to its MRO.
             abstract_bases.append(base)
@@ -84,12 +76,17 @@ def _c3_mro(cls, abcs=None):
     abstract_c3_mros = [_c3_mro(base, abcs=abcs) for base in abstract_bases]
     other_c3_mros = [_c3_mro(base, abcs=abcs) for base in other_bases]
     return _c3_merge(
-        [[cls]] +
-        explicit_c3_mros + abstract_c3_mros + other_c3_mros +
-        [explicit_bases] + [abstract_bases] + [other_bases]
+        [[cls]]
+        + explicit_c3_mros
+        + abstract_c3_mros
+        + other_c3_mros
+        + [explicit_bases]
+        + [abstract_bases]
+        + [other_bases]
     )
 
-def _compose_mro(cls, types):
+
+def _compose_mro(cls, types):  # noqa: C901
     """Calculates the method resolution order for a given class *cls*.
 
     Includes relevant abstract base classes (with their respective bases) from
@@ -97,11 +94,13 @@ def _compose_mro(cls, types):
 
     """
     bases = set(cls.__mro__)
+
     # Remove entries which are already present in the __mro__ or unrelated.
     def is_related(typ):
-        return (typ not in bases and hasattr(typ, '__mro__')
-                                 and issubclass(cls, typ))
+        return typ not in bases and hasattr(typ, '__mro__') and issubclass(cls, typ)
+
     types = [n for n in types if is_related(n)]
+
     # Remove entries which are strict bases of other entries (they will end up
     # in the MRO anyway.
     def is_strict_base(typ):
@@ -109,6 +108,7 @@ def _compose_mro(cls, types):
             if typ != other and typ in other.__mro__:
                 return True
         return False
+
     types = [n for n in types if not is_strict_base(n)]
     # Subclasses of the ABCs in *types* which are also implemented by
     # *cls* can be used to stabilize ABC ordering.
@@ -116,7 +116,7 @@ def _compose_mro(cls, types):
     mro = []
     for typ in types:
         found = []
-        for sub in filter(_safe, typ.__subclasses__()):
+        for sub in typ.__subclasses__():
             if sub not in bases and issubclass(cls, sub):
                 found.append([s for s in sub.__mro__ if s in type_set])
         if not found:
@@ -131,13 +131,6 @@ def _compose_mro(cls, types):
     return _c3_mro(cls, abcs=mro)
 
 
-def _safe(class_):
-    """
-    Return if the class is safe for testing as subclass. Ref #2.
-    """
-    return not getattr(class_, '__origin__', None)
-
-
 def _find_impl(cls, registry):
     """Returns the best matching implementation from *registry* for type *cls*.
 
@@ -154,17 +147,20 @@ def _find_impl(cls, registry):
         if match is not None:
             # If *match* is an implicit ABC but there is another unrelated,
             # equally matching implicit ABC, refuse the temptation to guess.
-            if (t in registry and t not in cls.__mro__
-                              and match not in cls.__mro__
-                              and not issubclass(match, t)):
-                raise RuntimeError("Ambiguous dispatch: {0} or {1}".format(
-                    match, t))
+            if (
+                t in registry
+                and t not in cls.__mro__
+                and match not in cls.__mro__
+                and not issubclass(match, t)
+            ):
+                raise RuntimeError(f"Ambiguous dispatch: {match} or {t}")
             break
         if t in registry:
             match = t
     return registry.get(match)
 
-def singledispatch(func):
+
+def singledispatch(func):  # noqa: C901
     """Single-dispatch generic function decorator.
 
     Transforms a function into a generic function, which can have different
@@ -172,12 +168,15 @@ def singledispatch(func):
     function acts as the default implementation, and additional
     implementations can be registered using the register() attribute of the
     generic function.
-
     """
+    # There are many programs that use functools without singledispatch, so we
+    # trade-off making singledispatch marginally slower for the benefit of
+    # making start-up of such applications slightly faster.
+    import types, weakref  # noqa: E401
+
     registry = {}
-    dispatch_cache = WeakKeyDictionary()
-    def ns(): pass
-    ns.cache_token = None
+    dispatch_cache = weakref.WeakKeyDictionary()
+    cache_token = None
 
     def dispatch(cls):
         """generic_func.dispatch(cls) -> <function implementation>
@@ -186,11 +185,12 @@ def singledispatch(func):
         for the given *cls* registered on *generic_func*.
 
         """
-        if ns.cache_token is not None:
+        nonlocal cache_token
+        if cache_token is not None:
             current_token = get_cache_token()
-            if ns.cache_token != current_token:
+            if cache_token != current_token:
                 dispatch_cache.clear()
-                ns.cache_token = current_token
+                cache_token = current_token
         try:
             impl = dispatch_cache[cls]
         except KeyError:
@@ -207,38 +207,81 @@ def singledispatch(func):
         Registers a new implementation for the given *cls* on a *generic_func*.
 
         """
+        nonlocal cache_token
         if func is None:
             if isinstance(cls, type):
                 return lambda f: register(cls, f)
             ann = getattr(cls, '__annotations__', {})
             if not ann:
                 raise TypeError(
-                    "Invalid first argument to `register()`: {cls!r}. "
-                    "Use either `@register(some_class)` or plain `@register` "
-                    "on an annotated function.".format(**locals())
+                    f"Invalid first argument to `register()`: {cls!r}. "
+                    f"Use either `@register(some_class)` or plain `@register` "
+                    f"on an annotated function."
                 )
             func = cls
 
+            # only import typing if annotation parsing is necessary
+            from typing import get_type_hints
+
             argname, cls = next(iter(get_type_hints(func).items()))
             if not isinstance(cls, type):
                 raise TypeError(
-                    "Invalid annotation for {argname!r}. "
-                    "{cls!r} is not a class.".format(**locals())
+                    f"Invalid annotation for {argname!r}. " f"{cls!r} is not a class."
                 )
         registry[cls] = func
-        if ns.cache_token is None and hasattr(cls, '__abstractmethods__'):
-            ns.cache_token = get_cache_token()
+        if cache_token is None and hasattr(cls, '__abstractmethods__'):
+            cache_token = get_cache_token()
         dispatch_cache.clear()
         return func
 
     def wrapper(*args, **kw):
+        if not args:
+            raise TypeError(f'{funcname} requires at least ' '1 positional argument')
+
         return dispatch(args[0].__class__)(*args, **kw)
 
+    funcname = getattr(func, '__name__', 'singledispatch function')
     registry[object] = func
     wrapper.register = register
     wrapper.dispatch = dispatch
-    wrapper.registry = MappingProxyType(registry)
+    wrapper.registry = types.MappingProxyType(registry)
     wrapper._clear_cache = dispatch_cache.clear
     update_wrapper(wrapper, func)
     return wrapper
 
+
+# Descriptor version
+class singledispatchmethod:
+    """Single-dispatch generic method descriptor.
+
+    Supports wrapping existing descriptors and handles non-descriptor
+    callables as instance methods.
+    """
+
+    def __init__(self, func):
+        if not callable(func) and not hasattr(func, "__get__"):
+            raise TypeError(f"{func!r} is not callable or a descriptor")
+
+        self.dispatcher = singledispatch(func)
+        self.func = func
+
+    def register(self, cls, method=None):
+        """generic_method.register(cls, func) -> func
+
+        Registers a new implementation for the given *cls* on a *generic_method*.
+        """
+        return self.dispatcher.register(cls, func=method)
+
+    def __get__(self, obj, cls=None):
+        def _method(*args, **kwargs):
+            method = self.dispatcher.dispatch(args[0].__class__)
+            return method.__get__(obj, cls)(*args, **kwargs)
+
+        _method.__isabstractmethod__ = self.__isabstractmethod__
+        _method.register = self.register
+        update_wrapper(_method, self.func)
+        return _method
+
+    @property
+    def __isabstractmethod__(self):
+        return getattr(self.func, '__isabstractmethod__', False)
